/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* $Id: minrtsFrameData.c,v 1.44 2004/10/07 13:29:09 jdejong Exp $ */
/*----------------------------------------------------------------------- */
/* Structure and methods for MIDI frame data                              */
/* This is related to imaging_data_row structures                         */
/*----------------------------------------------------------------------- */
#include "minrtsFrameData.h"
#include <math.h>
#include <stdio.h>

/* total size of MIDI detector frames */
#define MIDI_FRAME_NX 320
#define MIDI_FRAME_NY 240

/**
 *   Make an minrtsFrameData structure from an imaging_detector_table_def.
 *   Use minrtsFrameDataKill to destroy the result when not needed.
 *   Inputs:
 *    \param   detect     pointer to imaging_data table structure
 *   Returns:
 *      An minrtsFrameData structure pointer; NULL indicates failure
 */
minrtsFrameData* minrtsFrameDataMake (imaging_detector_table_def* detect)
{
  minrtsFrameData *out;
  int i, nregion, *naxis;
  long nval;

  /* sanity check */
  if (!detect) return NULL;

  /* create basic structure */
  out = (minrtsFrameData*)malloc(sizeof(minrtsFrameData));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(minrtsFrameData));
  
  nregion = detect->nregion;
  out->nregion = detect->nregion;

  /* allocate arrays */
  out->opt_train = (int*)malloc(detect->maxtel*sizeof(int));
  memset (out->opt_train, 0, detect->maxtel*sizeof(int));
  out->targtype = (char*)malloc(nregion*sizeof(char));
  memset (out->targtype, 'U', nregion*sizeof(char));
  out->target = (int*)malloc(nregion*sizeof(int));
  memset (out->target, 0, nregion*sizeof(int));
  out->nx = (int*)malloc(nregion*sizeof(int));
  memset (out->nx, 0, nregion*sizeof(int));
  out->ny = (int*)malloc(nregion*sizeof(int));
  memset (out->ny, 0, nregion*sizeof(int));
  out->x0 = (int*)malloc(nregion*sizeof(int));
  memset (out->x0, 0, nregion*sizeof(int));
  out->y0 = (int*)malloc(nregion*sizeof(int));
  memset (out->y0, 0, nregion*sizeof(int));
  out->mean = (float*)malloc(nregion*sizeof(float));
  memset (out->mean, 0, nregion*sizeof(float));
  out->rms = (float*)malloc(nregion*sizeof(float));
  memset (out->rms, 0, nregion*sizeof(float));
  /* set initial weight to zero */
  out->weight = 0.0;
  out->rms[0] = -1.0;
  /* data arrays */
  out->data     = (float**)malloc(nregion*sizeof(float*));
  for (i=0; i<detect->nregion; i++) {
    naxis = detect->naxis[i];
    out->nx[i] = naxis[0];
    out->ny[i] = naxis[1];
    out->x0[i] = detect->corner[i][0];
    out->y0[i] = detect->corner[i][1];
    nval = out->nx[i] * out->ny[i];
    out->data[i] = (float*)malloc(nval*sizeof(float));
    memset (out->data[i], 0, nval*sizeof(float));
  }

  return out; /* done */
} /* end minrtsFrameDataMake */

/**
 *   Delete an minrtsFrameData structure.
 *   Inputs:
 *     \param  data   pointer to the minrtsFrameData structure
 */
void minrtsFrameDataKill (minrtsFrameData *data)
{
  int i;

  if (data==NULL) return; /* sanity check */
  if (data->title!=NULL) free(data->title);
  if (data->opt_train!=NULL) free(data->opt_train);
  if (data->targtype!=NULL) free(data->targtype);
  if (data->nx!=NULL) free(data->nx);
  if (data->ny!=NULL) free(data->ny);
  if (data->x0!=NULL) free(data->x0);
  if (data->y0!=NULL) free(data->y0);
  if (data->mean!=NULL) free(data->mean);
  if (data->rms!=NULL) free(data->rms);
  if (data->data!=NULL) {
    for (i=0; i<data->nregion; i++) 
      if (data->data[i]!=NULL) free(data->data[i]);
    free(data->data);
  }
  free (data);      /* basic structure */
} /* end minrtsFrameDataKill */


/***************************************************************
 * Read through all raw files reading separate target and sky frames 
 * and accumulating them in given minrtsFrameData structures; 
 * Normalize by number of frames.
 * If there are no off-source measurements, offData is zero filled.
 * Inputs:
 *    fileList     array of file names of raw MIDI data
 *    nFiles       number of file names in fileList
 *    nDrop        drop this many points after change of chop phase
 *    onData       FrameData to receive average on-source data
 *    offData      FrameData to receive average off-source data
 *    ierr         error code, 0=>OK
 **************************************************************** */
void minrtsFrameDataUnChop(char **fileList, int nFiles, int nDrop,
			   minrtsFrameData *onFData, minrtsFrameData *offFData,
			   int *ierr)
{
   fitsfile *inFile;
   imaging_data_table_def *idata;
   imaging_data_row *inData;
   imaging_detector_table_def *detector;
   int iFile;

   long iRow, nRow;
   int i, iReg;
   int TN, SN;
   long iPix;
   int sinceChange = 0;
   char Sky = 'S';
   char Target = 'T';
   char targ, oldTarg;
   float *onData, *offData;
   float normSky, normTarget;
   int npix, OK;

   *ierr    = 0;
   TN = 0;
   SN = 0;
   oldTarg = ' ';

   /* Check that two FrameDatas are compatable */
   OK = onFData->nregion == offFData->nregion;
   for (iReg=0; iReg<onFData->nregion; iReg++) {
     OK = OK && (onFData->nx[iReg] == offFData->nx[iReg]);
     OK = OK && (onFData->ny[iReg] == offFData->ny[iReg]);
   }
   if (!OK) {
     fprintf (stderr," minrtsFrameDataUnChop: input FrameDatas are incompatable\n");
     *ierr = 1;
     return;
   }
   

   /* Initialize output arrays */
   for (iReg=0; iReg<onFData->nregion; iReg++) {
     npix = onFData->nx[iReg]*onFData->ny[iReg];
     onData  = onFData->data[iReg];
     offData = offFData->data[iReg];
     for (iPix=0;iPix<npix;iPix++) {
       onData[iPix] = 0.0;
       offData[iPix] = 0.0;
     }
   }

   /*  loop over files, read rows, read target types, add data
       into appropriate accumulators 
   */

   for (iFile =0; iFile <nFiles; iFile++) {
     /*   open file, open table, create table row */
      i = fits_open_file(&inFile, fileList[iFile], READONLY, ierr);
      if (*ierr !=0) {
         fprintf(stderr,"minrtsFrameDataUnChop: error %i opening file %s\n", 
		*ierr, fileList[iFile]);
	 fits_report_error(stderr, *ierr);
         return;
      }

      /* open table */
      idata = open_imaging_data_table(inFile, ierr);
      if (*ierr !=0) {
         fprintf(stderr,"minrtsFrameDataUnChop: error %i opening imaging data in file %s\n", 
            *ierr, fileList[iFile]);
	 fits_report_error(stderr, *ierr);
         return;
      }

      /* Check that input FrameData compatable with file */
      detector = idata->detector;
      OK = onFData->nregion == detector->nregion;
	for (iReg=0; iReg<onFData->nregion; iReg++) {
	  OK = OK && (onFData->nx[iReg] == detector->naxis[iReg][0]);
	  OK = OK && (onFData->ny[iReg] == detector->naxis[iReg][1]);
	}
      if (!OK) {
	fprintf (stderr,"minrtsFrameDataUnChop: input FrameData incompatable with file %s\n",
		 fileList[iFile]);
	*ierr = 1;
	return;
      }

      inData = make_imaging_data_row(idata);
      i = fits_get_num_rows (inFile,  &nRow, ierr);
      /* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++)
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, ierr);

      /* read lines, check target type and decide whether to accumulate */
      for (iRow =1; iRow<=nRow; iRow++) {
         i = read_imaging_data_table(inFile, idata, inData, iRow, ierr);
         if (*ierr !=0) {
            fprintf(stderr,"minrtsFrameDataUnChop: error %i reading imaging data line %ld in file %s\n", 
               *ierr, iRow, fileList[iFile]);
	    fits_report_error(stderr, *ierr);
	    return;
         }
	 /* See if any region has 'S' or 'T' as target type */
	 targ = ' ';
	 for (iReg=0; iReg<onFData->nregion; iReg++) {
            if ((inData->targtype[iReg]=='T') || (inData->targtype[iReg]=='S')) 
		targ = inData->targtype[iReg];
	 }

         if (targ != oldTarg ) {
            sinceChange = 0;
            oldTarg     = targ;
         } else {
            ++sinceChange;
         }

	 /*  if sinceChange > nDrop decide what type and accumulate data */
         if (sinceChange >= nDrop && (targ == Target)) {
	 /* Target data */
	 for (iReg=0; iReg<onFData->nregion; iReg++) {
	   npix = onFData->nx[iReg]*onFData->ny[iReg];
	   onData  = onFData->data[iReg];
  	   for (iPix=0; iPix<npix; iPix++) {
	     onData[iPix] += 32768. + (short)inData->data[iReg][iPix];
	   }
	 }
            ++TN;
         } else if (sinceChange >= nDrop && (targ == Sky)) {
	 /* Sky data */
	 for (iReg=0; iReg<onFData->nregion; iReg++) {
	   npix = offFData->nx[iReg]*offFData->ny[iReg];
	   offData = offFData->data[iReg];
  	   for (iPix=0; iPix<npix; iPix++) {
	     offData[iPix] += 32768. + (short)inData->data[iReg][iPix];
	   }
	 }
            ++SN;
         }
      } /* read rows in one file */
      
      /* close file and table, destroy row */
      i= close_imaging_data_table(inFile, idata, ierr);
      kill_imaging_data_row(inData);
   } /* end read all files in list */

   /* Normalize output arrays */
   if (TN>0) normTarget = 1.0 / (float)TN;
   else normTarget = 1.0;
   if (SN>0) normSky = 1.0 / (float)SN;
   else normSky = 0.0;
   for (iReg=0; iReg<onFData->nregion; iReg++) {
     npix = onFData->nx[iReg]*onFData->ny[iReg];
     onData  = onFData->data[iReg];
     offData = offFData->data[iReg];
     for (iPix=0; iPix<npix; iPix++) {
       onData[iPix]  *= normTarget;
       offData[iPix] *= normSky;
     }
   }

   /* Save total number of frames as weight */
   onFData->weight  = TN;
   offFData->weight = SN;

   /* Set other info */
   for (iReg=0; iReg<onFData->nregion; iReg++) {
     onFData->targtype[iReg]  = Target;
     offFData->targtype[iReg] = Sky;
   }

   return;
} /* end minrtsFrameDataUnChop */

/**
 *   Average a set of Frames in the form of imaging_data_table_rows
 *   producing a  minrtsFrameData.
 *   Descriptive information is copied from the first frame.
 *   Use minrtsFrameDataKill to destroy the result when not needed.
 *   Inputs:
 *   \param    detect    structure describing detector layout
 *   \param    nframes   Number of frames to be averaged
 *   \param    frames    Frame data to be averaged
 *   Returns:
 *   An minrtsFrameData pointer, NULL indicates failure
 */
minrtsFrameData* minrtsFrameDataAverage (imaging_detector_table_def* detect,
					 int nframes,
					 imaging_data_row** frames)
{
  minrtsFrameData *out;
  int reg, frame, i, len;
  DataType *inp;
  float    *outp, norm;

  /* sanity checks */
  if (!detect) return NULL;
  if (!frames) return NULL;
  if (nframes <=0) return NULL;

  /* create and initialize output */
  out = minrtsFrameDataMake (detect);
  if (!out) return out;

  /* copy descriptive info from first frame */
  out->time      = frames[0]->time;
  out->opd[0] = frames[0]->opd[0];
  out->opd[1] = frames[0]->opd[1];
  out->local_opd[0] = frames[0]->local_opd[0];
  out->local_opd[1] = frames[0]->local_opd[1];
  out->exp       = frames[0]->exp;
  out->offset[0] = frames[0]->offset[0];
  out->offset[1] = frames[0]->offset[1];
  out->rotation  = frames[0]->rotation;
  out->frame     = frames[0]->frame;
  out->reference = frames[0]->reference ;
  out->step      = frames[0]->step;
  out->nregion   = frames[0]->nregion;
  out->opt_train[0] = frames[0]->opt_train[0];
  out->opt_train[1] = frames[0]->opt_train[1];

  for (reg=0; reg<out->nregion; reg++)
    {
      out->nx[reg]        = detect->naxis[reg][0];
      out->ny[reg]        = detect->naxis[reg][1];
      out->targtype[reg]  = frames[0]->targtype[reg];
      out->target[reg]    = frames[0]->target[reg];
    }

  /* loop over frames summing, values initially zero */
  for (frame=0; frame<nframes; frame++) 
    {
      /* loop over regions */
      for (reg=0; reg<out->nregion; reg++)
	{
	  inp  = frames[frame]->data[reg];
	  outp = out->data[reg];
	  len = out->nx[reg] * out->ny[reg]; /* how many elements */
	  for (i = 0; i<len; i++) 
	    {
	      outp[i] += (float)inp[i];
	    }
	} /* end loop over regions */
    } /* end loop over frames */

  /* normalize to get average */
  norm = 1.0 / (float)nframes;
  /* set weight */
  out->weight = (float)nframes;
  for (reg=0; reg<out->nregion; reg++)
    {
      outp = out->data[reg];
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      for (i = 0; i<len; i++) 
	{
	  outp[i] *= norm;
	}
    } /* end loop over regions */
  return out;
} /* end minrtsFrameDataAverage */

/**
 *   Get statistics for each region and fill in mean and rms entries.
 *   Iterate a couple times to discard outliers.
 *   Inputs:
 *    \param   data      minrtsFrameData to operate on
 */
void minrtsFrameDataStats (minrtsFrameData* data)
{
  long reg, i, len, count, iter;
  double sum, sum2;
  float  *values, sigma, mean, limit;

  /* sanity check */
  if (!data) return;

  /* loop over regions */
  for (reg = 0; reg < data->nregion; reg++)
    {
      sigma = 1.0e20;  /* something big the first time */
      mean = 0.0;
      /* iterate dropping outliers > 5 sigma */
      for (iter=0; iter<3; iter++) 
	{
	  limit = 5.0 * sigma; /* highest value acceptable */
	  count = 0; sum = 0.0; sum2 = 0.0;
	  len = data->nx[reg] * data->ny[reg]; /* how many elements */
	  values = data->data[reg];
	  /* over array */
	  for (i = 0; i<len; i++) 
	    {
	      if (abs (values[i]-mean) < limit) /* want this one? */
		{
		  count = count + 1;
		  sum = sum + values[i];
		  sum2 = sum2 + values[i]*values[i];
		}
	    }
	  /* save results in structure */
	  if (count > 1) /* need enough for statistics */
	    {
	      sum = sum / count;
	      sum2 = sum2 / count;
	      mean = sum;
	      data->mean[reg] = mean;
	      sigma = sqrt (sum2 - sum*sum);
	      data->rms[reg] = sigma;
	    } else { /* too little data */
	      data->mean[reg] = -1.0;
	      data->rms[reg]  = -1.0;
	    }
	} /* end of statistics iteration loop */
    } /* end loop over region */
} /* end minrtsFrameDataStats */

/**
 *   Convert minrtsFrameData to an image of the detector.
 *   Inputs
 *    \param   data      minrtsFrameData to operate on
 *   Output:
 *     \param  nx        Number of columns in result
 *     \param  ny        Number of rows in result
 *   Returns:
 *      An array of pixel values; NULL indicates failure
 *      pixels are ordered by x then y
 *      zeroes are used for pixels not in a region
 */
float* minrtsFrameDataImage (minrtsFrameData* data, 
			     int *nx, int *ny)
{
  float *out, *inp;
  long  size, reg, mx, my, ix, iy, iIn, iOut, xCorn, yCorn;

  /* sanity checks */
  if (!data)   return NULL;
  if (!nx)     return NULL;
  if (!ny)     return NULL;

  /* get size of output */
  *nx = MIDI_FRAME_NX;
  *ny = MIDI_FRAME_NY;

  /* create output */
  size = MIDI_FRAME_NX * MIDI_FRAME_NY;
  out = (float*)malloc(size*sizeof(float));
  if (!out) return out; /* allocation error check */
  memset (out, 0, size*sizeof(float)); /* zero fill */

  /* loop over regions copying values to relevant positions */
  for (reg=0; reg<data->nregion; reg++)
    {
      /* corner in MIDI frame */
      xCorn = data->x0[reg];
      yCorn = data->y0[reg];

      /* size of region */
      mx = data->nx[reg];
      my = data->ny[reg];

      /* make sure they don't exceed image border */
      if (xCorn+mx>(*nx)) mx = (*nx) - xCorn;
      if (yCorn+my>(*ny)) my = (*ny) - yCorn;

      /* region data pointer */
      inp = data->data[reg];

      /* loops over axes of region */
      for (iy=0; iy<my; iy++)
	{
	  /* index in image */
	  iOut = ((yCorn + iy) * (*nx)) + xCorn; 
	  /* index in region data */
	  iIn = iy * mx;  
	  for (ix=0; ix<mx; ix++) /* copy */
	    {
	      out[iOut++] = inp[iIn++];
	    }
	}
    } /* end loop over region */

  return out; /* done */
} /* end minrtsFrameDataImage  */


/**
 *   Add a single image row to an existing FrameData object
 *   Inputs:
 *    \param   data    existing FrameData object
 *    \param   row     image data row  
 *
 */
void minrtsFrameDataRowAdd (minrtsFrameData* data,
	                   imaging_data_row* row)
{
  int reg, i, len;
  float *valIn;
  DataType *rowData;

  /* sanity checks */
  if (!data) return;
  if (!row) return;

  /* loop over regions */
  for (reg = 0; reg < data->nregion; reg++)
    {
      len = data->nx[reg] * data->ny[reg]; /* how many elements */
      valIn    = data->data[reg];
      rowData  = row->data[reg];
      for (i = 0; i<len; i++) 
        {
           valIn[i] = valIn[i] + (float)rowData[i];
        } /* end loop over frames */
     } /* end loop over region */
   /* increment weights */
   data->weight = data->weight + 1;
} /* end minrtsFrameDataRowAdd */

/**
 *   Pixel by pixel out = in1 + in2
 *   Inputs:
 *    \param   in1       first minrtsFrameData operand
 *    \param   in2       second minrtsFrameData operand
 *   Output:
 *    \param   out       output minrtsFrameData (must exist prior to call)
 */
void minrtsFrameDataAdd (minrtsFrameData* out,
			 minrtsFrameData* in1,
			 minrtsFrameData* in2)
{
  int reg, i, len;
  float *valIn1, *valIn2, *valOut;

  /* sanity checks */
  if (!out) return;
  if (!in1) return;
  if (!in2) return;

  /* loop over regions */
  for (reg = 0; reg < out->nregion; reg++)
    {
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      valIn1 = in1->data[reg];
      valIn2 = in2->data[reg];
      valOut = out->data[reg];
      for (i = 0; i<len; i++) 
	{
	  valOut[i] = valIn1[i] + valIn2[i];
	}
     } /* end loop over region */
   /* sum weights */
   out->weight = in1->weight + in2->weight;
} /* end minrtsFrameDataAdd */

/**
 *   Pixel by pixel out = in1 - in2
 *   Inputs:
 *    \param   in1       first minrtsFrameData operand
 *    \param   in2       second minrtsFrameData operand
 *   Output:
 *    \param   out       output minrtsFrameData (must exist prior to call)
 */
void minrtsFrameDataSub (minrtsFrameData* out,
			 minrtsFrameData* in1,
			 minrtsFrameData* in2)
{
  int reg, i, len;
  float *valIn1, *valIn2, *valOut;

  /* sanity checks */
  if (!out) return;
  if (!in1) return;
  if (!in2) return;

  /* loop over regions */
  for (reg = 0; reg < out->nregion; reg++)
    {
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      valIn1 = in1->data[reg];
      valIn2 = in2->data[reg];
      valOut = out->data[reg];
      for (i = 0; i<len; i++) 
	{
	  valOut[i] = valIn1[i] - valIn2[i];
	}
     } /* end loop over region */
     out->weight = in1->weight + in2->weight;
} /* end minrtsFrameDataSub */

/**
 *   Pixel by pixel out = in1 / in2
 *   Inputs:
 *    \param   in1       first minrtsFrameData operand
 *    \param   in2       second minrtsFrameData operand
 *   Output:
 *    \param   out       output minrtsFrameData (must exist prior to call)
 */
void minrtsFrameDataDiv (minrtsFrameData* out,
			 minrtsFrameData* in1,
			 minrtsFrameData* in2)
{
  int reg, i, len;
  float *valIn1, *valIn2, *valOut;

  /* sanity checks */
  if (!out) return;
  if (!in1) return;
  if (!in2) return;

  /* loop over regions */
  for (reg = 0; reg < out->nregion; reg++)
    {
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      valIn1 = in1->data[reg];
      valIn2 = in2->data[reg];
      valOut = out->data[reg];
      for (i = 0; i<len; i++) 
	{
	  valOut[i] = valIn1[i] / valIn2[i];
	}
     } /* end loop over region */
     out->weight = 1.0;
} /* end minrtsFrameDataDiv */

/**
 *   Pixel by pixel out = in1 * in2
 *   Inputs:
 *    \param   in1       first minrtsFrameData operand
 *    \param   in2       second minrtsFrameData operand
 *   Output:
 *     \param  out       output minrtsFrameData (must exist prior to call)
 */
void minrtsFrameDataMul (minrtsFrameData* out,
			 minrtsFrameData* in1,
			 minrtsFrameData* in2)
{
  int reg, i, len;
  float *valIn1, *valIn2, *valOut;

  /* sanity checks */
  if (!out) return;
  if (!in1) return;
  if (!in2) return;

  /* loop over regions */
  for (reg = 0; reg < out->nregion; reg++)
    {
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      valIn1 = in1->data[reg];
      valIn2 = in2->data[reg];
      valOut = out->data[reg];
      for (i = 0; i<len; i++) 
	{
	  valOut[i] = valIn1[i] * valIn2[i];
	}
     } /* end loop over region */
     out->weight = 1.0;
} /* end minrtsFrameDataMul */


void minrtsFrameDataRMS (minrtsFrameData* out,
			 minrtsFrameData* in1,
			 minrtsFrameData* in2)
{
  int reg, i, len;
  float *valIn1, *valIn2, *valOut;

  /* sanity checks */
  if (!out) return;
  if (!in1) return;
  if (!in2) return;

  /* loop over regions */
  for (reg = 0; reg < out->nregion; reg++)
    {
      len = out->nx[reg] * out->ny[reg]; /* how many elements */
      valIn1 = in1->data[reg];
      valIn2 = in2->data[reg];
      valOut = out->data[reg];
      for (i = 0; i<len; i++) 
	{
	  valOut[i] = sqrt(valIn1[i] - valIn2[i]*valIn2[i]);
	}
     } /* end loop over region */
} /* end minrtsFrameDataRMS */

void minrtsFrameDataScalarMul (minrtsFrameData* in, float factor) {
  int reg, i, len;
  float *valIn;

  /* sanity checks */
  if (!in) return;

  /* loop over regions */
  for (reg = 0; reg < in->nregion; reg++) {
     len = in->nx[reg] * in->ny[reg]; /* how many elements */
     valIn = in->data[reg];
     for (i = 0; i<len; i++) valIn[i] = valIn[i] * factor;
     }
  return;
} /* end minrtsFrameDataScalarMul */

void minrtsFrameDataScalarAdd (minrtsFrameData* in, float add) {
  int reg, i, len;
  float *valIn;

  /* sanity checks */
  if (!in) return;

  /* loop over regions */
  for (reg = 0; reg < in->nregion; reg++) {
     len = in->nx[reg] * in->ny[reg]; /* how many elements */
     valIn = in->data[reg];
     for (i = 0; i<len; i++) valIn[i] = valIn[i] + add;
     }
  return;
} /* end minrtsFrameDataScalarAdd */

/*************************************************************
  * In place square root, set <= 0 to 0.
  ******************************************************** */
void minrtsFrameDataSqrt (minrtsFrameData* in) {
  int reg, i, len;
  float *valIn;

  /* sanity checks */
  if (!in) return;

  /* loop over regions */
  for (reg = 0; reg < in->nregion; reg++) {
     len = in->nx[reg] * in->ny[reg]; /* how many elements */
     valIn = in->data[reg];
     for (i = 0; i<len; i++) 
        if (valIn[i] > 0.) valIn[i] = sqrt(valIn[i]); else
           valIn[i] = 0.;
     }
  return;
} /* end minrtsFrameDataSqrt */
   /**  
     * duplicate an existing FrameData structure
     * Inputs:
     * \param   in     pointer to minrtsFrameData   target of insertion
     *  Outputs:       pointer to minrtsFrameData   data from specified region
   */
minrtsFrameData *minrtsFrameDataClone (minrtsFrameData *in){
   /* local variables */
   int i, nregion;
   long nData, j;
   minrtsFrameData *outFrame = NULL;
   outFrame = (minrtsFrameData *)malloc (sizeof(minrtsFrameData));
/* copy the easy stuf */
   outFrame->title = NULL;
   outFrame->time = in->time;
   outFrame->opd[0] = in->opd[0];
   outFrame->opd[1] = in->opd[1];
   outFrame->local_opd[1] = in->local_opd[1];
   outFrame->local_opd[0] = in->local_opd[0];
   outFrame->exp = in->exp;
   outFrame->offset[0] = in->offset[0];
   outFrame->offset[1] = in->offset[1];
   outFrame->rotation = in->rotation;
   outFrame->frame = in->frame;
   outFrame->reference = in->reference;
   outFrame->step = in->step;
   outFrame->weight = in->weight;
   outFrame->nregion = in->nregion;
   nregion = in->nregion;
/* now we have to allocate space */
   outFrame->nx = (int *)malloc(nregion * sizeof(int));
   outFrame->ny = (int *)malloc(nregion * sizeof(int));
   outFrame->x0 = (int *)malloc(nregion * sizeof(int));
   outFrame->y0 = (int *)malloc(nregion * sizeof(int));
   outFrame->opt_train = (int *)malloc(2 * sizeof(int));
   outFrame->targtype = (char*)malloc(nregion * sizeof(char));
   outFrame->target = (int *)malloc(nregion * sizeof(int));
   outFrame->mean = (float *)malloc(nregion * sizeof(float));
   outFrame->rms = (float *)malloc(nregion * sizeof(float));
   outFrame->data = (float **)malloc(nregion * sizeof(float*));
   
   for (i=0; i<2; i++)
      outFrame->opt_train[i] = in->opt_train[i];

   for (i=0; i<nregion; i++) {
      outFrame->nx[i] = in->nx[i];
      outFrame->ny[i] = in->ny[i];
      outFrame->x0[i] = in->x0[i];
      outFrame->y0[i] = in->y0[i];
      nData = in->nx[i] * (long)in->ny[i];
      outFrame->targtype[i] = in->targtype[i];
      outFrame->target[i] = in->target[i];
      outFrame->mean[i] = in->mean[i];
      outFrame->rms[i] = in->rms[i];
      outFrame->data[i] = (float*)malloc(nData * sizeof(float));
      for (j=0; j<nData; j++) outFrame->data[i][j] = in->data[i][j];
   }
   return outFrame;
   }
/*   end of minrtsFrameDataClone  */

/**
  *    Write the contents of a minrtsFrameData structure out
  *    into a FITS table data on disk, setting many values
  *    to elementary defaults, but at least also copying
  *    the detector structure
  *    Inputs:
  *    \param  in          minrtsFrameData pointer  where data comes from
  *    \param fileName     char string              name of output FITS file
  *    \param ierr         pointer to int           if not 0 its not good
  *    Outputs:  none
  */
void minrtsFrameDataWriteFits (minrtsFrameData* in, char* fileName, 
   int *ierr) {
/*local variables */
   fitsfile *outFile = NULL;
   int i, nregion;
   long nData, iData;
   char **regname = NULL;
   imaging_detector_table_def *detector;
   imaging_fdata_row *dataRow;
   imaging_fdata_table_def *dataDef; 
   int *region; /*array of region numbers (1->nregion)*/
   int *detNo = NULL;
   int **ports = NULL;
   int *correlation;
   int **corner;
   double *gain = NULL;
   int **naxis;
   double **crval = NULL;
   float **crpix = NULL;
   char ***ctype = NULL;
   double **cd = NULL;
   double ***dmc = NULL;
   int ****dmp = NULL;
   if (*ierr > 0) return;  /*existing error conditions*/
/*fill region array*/
   nregion = in->nregion;
   region = (int*) malloc(nregion * sizeof(int));
   correlation = (int*) malloc(nregion * sizeof(int));
   for (i=0; i<nregion;i++) {region[i]=i+1;}
   for (i=0; i<nregion;i++) {correlation[i]=2;}
/*fill in corners and naxis */
   corner = (int**) malloc(nregion *sizeof(int*));
   naxis = (int**) malloc(nregion *sizeof(int*));
   for (i=0; i<nregion;i++) {
      corner[i] = (int*) malloc(2 * sizeof(int));
      naxis[i]  = (int*) malloc(2 * sizeof(int));
      }
   for (i=0; i<nregion;i++) {
      corner[i][0] = in->x0[i];
      corner[i][1] = in->y0[i];
      naxis[i][0]  = in->nx[i];
      naxis[i][1]  = in->ny[i];
   }
   remove (fileName); /* make sure that there is no previous file */
   fits_create_file (&outFile, fileName, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error opening fits output \n");
      return;
      }
/*create output detector */
   detector = make_imaging_detector_table_def (
      nregion, 1, 2, 2, 1, "MIDI", "xxx", 
      region, regname, detNo, ports,
      correlation,
      corner, gain, naxis,
      crval,
      crpix,
      ctype, cd, dmc, dmp);
/*create data table while writing detector to disk*/
   dataDef = make_imaging_fdata_table_def (
      outFile, detector, 20, ierr); /*fictitious maxstep of 20*/
   if (*ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return;
      }
/*template structure for a single row */
   dataRow = make_imaging_fdata_row(dataDef);
/*fill data row from in object */
   for (i=0; i<nregion;i++) {
      dataRow->targtype[i] = in->targtype[i];
      nData = in->nx[i] * in->ny[i];
      for (iData =0; iData<nData;iData++) dataRow->data[i][iData] = in->data[i][iData];
   }
/*write out row*/
   i = write_imaging_fdata_table(outFile, dataDef, dataRow, 1, ierr);
   i = close_imaging_fdata_table(outFile, dataDef, ierr);
   kill_imaging_fdata_row(dataRow);
   return;
} /* end of  miWindowFrameWriteFits */

/**
  *    Write the contents of an array of  minrtsFrameData structures 
  *    into a FITS table data on disk, setting many values
  *    to elementary defaults, but at least also copying
  *    the detector structure
  *    Inputs:
  *    \param  in          minrtsFrameData pointer  where data comes from
  *    \param fileName     char string              name of output FITS file
  *    \param ierr         pointer to int           if not 0 its not good
  *    Outputs:  none
  */
void minrtsFrameDataWriteFitsArray (minrtsFrameData** in, int nFrame,
   char* fileName, int *ierr) {
/*local variables */
   fitsfile *outFile = NULL;
   int i, nregion, iRow;
   long nData, iData;
   char **regname = NULL;
   imaging_detector_table_def *detector;
   imaging_fdata_row *dataRow;
   imaging_fdata_table_def *dataDef; 
   int *region; /*array of region numbers (1->nregion)*/
   int *detNo = NULL;
   int **ports = NULL;
   int *correlation;
   int **corner;
   double *gain = NULL;
   int **naxis;
   double **crval = NULL;
   float **crpix = NULL;
   char ***ctype = NULL;
   double **cd = NULL;
   double ***dmc = NULL;
   int ****dmp = NULL;
   if (*ierr > 0) return;  /*existing error conditions*/
/*fill region array*/
   nregion = in[0]->nregion;
   region = (int*) malloc(nregion * sizeof(int));
   correlation = (int*) malloc(nregion * sizeof(int));
   for (i=0; i<nregion;i++) {region[i]=i+1;}
   for (i=0; i<nregion;i++) {correlation[i]=2;}
/*fill in corners and naxis */
   corner = (int**) malloc(nregion *sizeof(int*));
   naxis = (int**) malloc(nregion *sizeof(int*));
   for (i=0; i<nregion;i++) {
      corner[i] = (int*) malloc(2 * sizeof(int));
      naxis[i]  = (int*) malloc(2 * sizeof(int));
      }
   for (i=0; i<nregion;i++) {
      corner[i][0] = in[0]->x0[i];
      corner[i][1] = in[0]->y0[i];
      naxis[i][0]  = in[0]->nx[i];
      naxis[i][1]  = in[0]->ny[i];
   }
   remove (fileName);
   fits_create_file (&outFile, fileName, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error opening fits output \n");
      return;
      }
/*create output detector */
   detector = make_imaging_detector_table_def (
      nregion, 1, 2, 2, 1, "MIDI", "xxx", 
      region, regname, detNo, ports,
      correlation,
      corner, gain, naxis,
      crval,
      crpix,
      ctype, cd, dmc, dmp);
/*create data table while writing detector to disk*/
   dataDef = make_imaging_fdata_table_def (
      outFile, detector, 20, ierr); /*fictitious maxstep of 20*/
   if (*ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return;
      }
/*template structure for a single row */
   dataRow = make_imaging_fdata_row(dataDef);
/*fill data row from in object */
   for (iRow =0; iRow < nFrame; iRow++) {
      for (i=0; i<nregion;i++) {
         dataRow->targtype[i] = in[iRow]->targtype[i];
         nData = in[iRow]->nx[i] * in[iRow]->ny[i];
         for (iData =0; iData<nData;iData++) dataRow->data[i][iData] = in[iRow]->data[i][iData];
      }
/*write out row*/
      i = write_imaging_fdata_table(outFile, dataDef, dataRow, 1+iRow, ierr);
   }  /* end of loop over rows */
   i = close_imaging_fdata_table(outFile, dataDef, ierr);
   kill_imaging_fdata_row(dataRow);
   return;
} /* end of  miWindowFrameWriteFits */

/**
  *    Read the contents of a minrtsFrameData structure 
  *    from a FITS table data on disk
  *    Inputs:
  *    \param fileName     char string              name of output FITS file
  *    \param row          int                      which row in input table
  *    \param ierr         pointer to int           if not 0 its not good
  *    Outputs:  frame     pointer to minrtsFrameData
  */
minrtsFrameData *minrtsFrameDataReadFitsRow (char* fileName, int inRow,
   int *ierr) {
/*local variables */
   fitsfile                *inFile = NULL;
   imaging_fdata_row       *row   = NULL;
   minrtsFrameData         *frame = NULL;
   imaging_detector_table_def *detector = NULL;
   imaging_fdata_table_def *fdata = NULL;
   int nregion, iregion, i;
/*pre-existing error*/
   if (*ierr !=0) return frame;
/* open input file */
   fits_open_file(&inFile, fileName, READONLY, ierr);
   if (*ierr == 0) fdata = open_imaging_fdata_table(inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error opening input file %s\n", fileName);
      return frame;
      }
/* read detector definition table */
   detector = read_imaging_detector_table(inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error reading input detector %s\n", fileName);
      return frame;
      }
/* read table definition */
   fdata = open_imaging_fdata_table (inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error reading data table in %s\n", fileName);
      return frame;
      }
/* make an empty row */
   row = make_imaging_fdata_row(fdata);
/* read 1st and assumed only row */
   i = read_imaging_fdata_table(inFile, fdata, row, (long) inRow, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error reading data row %i in %s\n", inRow, fileName);
      return frame;
      }
/* create an empty frame */
   frame = minrtsFrameDataMake (detector);
/*transfer data arrays */
   nregion = frame->nregion;
   for (iregion=0; iregion<nregion; iregion++){
      for (i=0; i< frame->nx[iregion]* (long)frame->ny[iregion]; i++)
         frame->data[iregion][i] =  row->data[iregion][i];
   }
/*  close input file */
   i = close_imaging_fdata_table(inFile, fdata, ierr);
   return frame;
}
/* end of minrtsFrameDataReadFitsRow  */
/*
   Read Frame Data from first row only of a Fits table
*/
minrtsFrameData *minrtsFrameDataReadFits (char* fileName, int *iErr) {
   return minrtsFrameDataReadFitsRow(fileName, (long) 1, iErr);
}  /*  end of read fits 1st row */


/**
  *    compute pixel by pixel mean and rms
  *    from a FITS table data on disk
  *    Inputs:
  *    \param fileName     char string              name of input FITS file
  *    \param mean         pointer to pointer to FrameData  output mean
  *    \param rms          pointer to pointer to FrameData  output rms
  *    \param ierr         pointer to int           if not 0 its not good
  */
void minrtsFrameDataFitsMeanRms (char *fileName, minrtsFrameData **mean, 
   minrtsFrameData **rms, int *ierr) {
/*local variables */
   fitsfile            *inFile = NULL;
   imaging_data_row    *row   = NULL;
   minrtsFrameData     *meanData = NULL;
   minrtsFrameData     *rmsData  = NULL;
   imaging_detector_table_def *detector = NULL;
   imaging_data_table_def *data = NULL;
   int nregion, iregion, i;
   long nRow, inRow;
   float x;
/*pre-existing error*/
   if (*ierr !=0) return ;
/* open input file */
   fits_open_file(&inFile, fileName, READONLY, ierr);
   if (*ierr == 0) data = open_imaging_data_table(inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error opening input file %s\n", fileName);
      return ;
      }
/* read detector definition table */
   detector = read_imaging_detector_table(inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error reading input detector %s\n", fileName);
      return ;
      }
   nregion=detector->nregion;
/* read table definition */
   data = open_imaging_data_table (inFile, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error reading data table in %s\n", fileName);
      return ;
      }
/* make an empty row */
   row = make_imaging_data_row(data);
/* number of rows */
   i = fits_get_num_rows (inFile,  &nRow, ierr);
/* suppress autoscaling to avoid overflow */
   for (i=0; i<data->nregion; i++)
      fits_set_tscale(inFile, 1+data->data_col[i], 1.0, 0.0, ierr);

/* create output frameData*/
   meanData = minrtsFrameDataMake (detector);
   rmsData  = minrtsFrameDataMake (detector);

/* read through data and sum */
   for (inRow=1; inRow <=nRow; inRow++) {
      i = read_imaging_data_table(inFile, data, row, inRow, ierr);
      if (*ierr !=0) {
         fprintf (stderr, "error reading data row %ld in %s\n", inRow, fileName);
         return ;
         }
/*sum data arrays */
      for (iregion=0; iregion<nregion; iregion++){
         for (i=0; i< meanData->nx[iregion]* (long)meanData->ny[iregion]; i++)
            meanData->data[iregion][i] +=  (short)row->data[iregion][i];
      }
   }
/*  divide by nRow */
      for (iregion=0; iregion<nregion; iregion++){
         for (i=0; i< meanData->nx[iregion]* (long)meanData->ny[iregion]; i++)
            meanData->data[iregion][i] = meanData->data[iregion][i]/(float)nRow;
      }
/*  go back and compute sum of (data - mean)^2
    we do this as a separate i/o operation to preserve accuracy
*/
   for (inRow=1; inRow <=nRow; inRow++) {
      i = read_imaging_data_table(inFile, data, row, inRow, ierr);
      if (*ierr !=0) {
         fprintf (stderr, "error reading data row %ld in %s\n", inRow, fileName);
         return ;
         }
      for (iregion=0; iregion<nregion; iregion++){
         for (i=0; i< meanData->nx[iregion]* (long)meanData->ny[iregion]; i++) {
            x = (short)row->data[iregion][i] - meanData->data[iregion][i];
            rmsData->data[iregion][i] +=  x*x;
         }
      }
   }
/*   take square root  */
      for (iregion=0; iregion<nregion; iregion++){
         for (i=0; i< meanData->nx[iregion]* (long)meanData->ny[iregion]; i++)
            rmsData->data[iregion][i] = sqrt(rmsData->data[iregion][i]/(float)nRow);
      }
/*    add bias to mean data */
      for (iregion=0; iregion<nregion; iregion++){
         for (i=0; i< meanData->nx[iregion]* (long)meanData->ny[iregion]; i++)
            meanData->data[iregion][i] += 32768.;
      }
/*  close input file */
   i = close_imaging_data_table(inFile, data, ierr);
   kill_imaging_data_row(row);
   kill_imaging_detector_table_def(detector);
   *mean = meanData;
   *rms  = rmsData;
   return ;
}  /* end of minrtsFrameDataFitsMeanRms */

   /**  
     * set float data from one of the
     * regions of a FrameData structure
     * Inputs:
     * \param   in     pointer to minrtsFrameData   target of insertion
     * \param   region int                          which region to modify, (1-rel)
     * \param   nx     int                          number of points in x-direction
     * \param   ny     int                          number of points in y-direction
     * \param   data   pointer to float             floating data
     * \param   ierr   pointer to int               if not 0, screwed up
     * Outputs         none
     */
void minrtsFrameDataSetData (minrtsFrameData* in, int region, 
   int nx, int ny, float* data, int *ierr) {
   long i;
   int iregion;
/*  check if region in range  */
   if ((region < 1) || (region > in->nregion)) {
      fprintf (stderr, "nregion out of range \n");
      *ierr = 1;
      return;
      }
    iregion = region -1; /* fits is 1-relative, c is not */
/*  check if size of array is OK   */
   if (nx*(long)ny != in->nx[iregion]*(long)in->ny[iregion]) {
      fprintf (stderr, "size of region doesnt match\n");
      *ierr = 1;
      return;
      }
      for (i=0; i< nx*(long)ny; i++) in->data[iregion][i] = data[i];
      *ierr = 0;
      return;
   } /* end of miWindowFrameSetData */

   /**  
     * get float data from one of the
     * regions of a FrameData structure
     * Inputs:
     * \param   in     pointer to minrtsFrameData   target of insertion
     * \param   region int                          which region to modify, (1-rel)
     * \param   ierr   pointer to int               if not 0, screwed up
     *  Outputs:       pointer to float             data from specified region
     */
float * minrtsFrameDataGetData (minrtsFrameData* in, int region, int *ierr) {
   long i, nData;
   int iregion;
   float * data = NULL;
/*  check if region in range  */
   if ((region < 1) || (region > in->nregion)) {
      fprintf (stderr, "nregion out of range \n");
      *ierr = 1;
      return (float *)NULL;
      }
      iregion = region - 1;
      nData = in->nx[iregion] * (long)in->ny[iregion];
      data = (float *)malloc(nData * sizeof(float));
      for (i=0; i< nData; i++) data[i] = in->data[iregion][i]; 
      *ierr = 0;
      return data;
   } /* end of miWindowFrameGetData */

   /**  
     * optionally multiply each region by a weight mask
     * and then sum all the pixels, to either in 0 or 1 dim
     * Inputs:
     * \param   in     pointer to minrtsFrameData   input frames
     * \param   mask   pointer to minrtsFrameData   mask frame
        can be NULL (no masking)
     * \param   nDim   int                          #dimension in result
     *  Outputs:       pointer to minrtsFrameData   compressed data 
     */
  minrtsFrameData * minrtsFrameDataCompress (minrtsFrameData *in, 
     minrtsFrameData *mask, int nDim) {
   /* local variables */
   int i, nregion, k;
   long nData, j;
   float sumData;
   float *inCol = NULL, *maskCol = NULL;
   minrtsFrameData *outFrame = NULL;
   outFrame = (minrtsFrameData *)malloc (sizeof(minrtsFrameData));
/* copy the easy stuf */
   outFrame->title = NULL;
   outFrame->time = in->time;
   outFrame->opd[0] = in->opd[0];
   outFrame->opd[1] = in->opd[1];
   outFrame->local_opd[1] = in->local_opd[1];
   outFrame->local_opd[0] = in->local_opd[0];
   outFrame->exp = in->exp;
   outFrame->offset[0] = in->offset[0];
   outFrame->offset[1] = in->offset[1];
   outFrame->rotation = in->rotation;
   outFrame->frame = in->frame;
   outFrame->reference = in->reference;
   outFrame->step = in->step;
   outFrame->weight = in->weight;
   outFrame->nregion = in->nregion;
   nregion = in->nregion;
/* now we have to allocate space */
   outFrame->nx = (int *)malloc(nregion * sizeof(int));
   outFrame->ny = (int *)malloc(nregion * sizeof(int));
   outFrame->x0 = (int *)malloc(nregion * sizeof(int));
   outFrame->y0 = (int *)malloc(nregion * sizeof(int));
   outFrame->opt_train = (int *)malloc(2 * sizeof(int));
   outFrame->targtype = (char*)malloc(nregion * sizeof(char));
   outFrame->target = (int *)malloc(nregion * sizeof(int));
   outFrame->mean = (float *)malloc(nregion * sizeof(float));
   outFrame->rms = (float *)malloc(nregion * sizeof(float));
   outFrame->data = (float **)malloc(nregion * sizeof(float*));

   for (i=0; i<2; i++)
      outFrame->opt_train[i] = in->opt_train[i];

/* if nDim != 1, assume it to be zero, i.e. reduce to single points */
   for (i=0; i<nregion; i++) {
      outFrame->ny[i] = 1;
      outFrame->y0[i] = 1+i;
      outFrame->x0[i] = in->x0[i];
      if (nDim == 1) outFrame->nx[i] = in->nx[i]; else outFrame->nx[i] = 1;
      nData = in->nx[i] * (long)in->ny[i];
      outFrame->targtype[i] = in->targtype[i];
      outFrame->target[i] = in->target[i];
      outFrame->mean[i] = in->mean[i];
      outFrame->rms[i] = in->rms[i];
      outFrame->data[i] = (float*)malloc(outFrame->nx[i] * sizeof(float));
/* actual compression  if Mask is NULL ignore it */
      if (nDim != 1) {
         sumData = 0.;
         if (mask == NULL) for (j=0; j<nData; j++) 
            sumData=sumData + in->data[i][j];
         else for (j=0;j<nData;j++) sumData = 
            sumData+(in->data[i][j])*(mask->data[i][j]);
         outFrame->data[i][0] = sumData;
      } else {
         for (k=0; k< (in->nx[i]);k++) {
            inCol = in->data[i]+k;
            if (mask != NULL) maskCol = mask->data[i]+k;
            sumData = 0.;
            if (mask == NULL) for (j=0; j<(in->ny[i]); j++) {
               sumData = sumData + *inCol;
               inCol = inCol + in->nx[i];
               }
            else for (j=0; j<(in->ny[i]);j++) {
               sumData = sumData + *inCol*(*maskCol);
               inCol   = inCol + in->nx[i];
               maskCol = maskCol + in->nx[i];
               }
            outFrame->data[i][k] = sumData;
         }
      }
   }
   return outFrame;
  }  /* end of minrtsFrameDataCompress */


/**  
  * Blank pixels where mask is zero or less.
  * Inputs:
  * \param   in     pointer to minrtsFrameData   input frame
  * \param   mask   pointer to minrtsFrameData   mask frame
  * \param   blank  blanking value
  */
void minrtsFrameDataBlank (minrtsFrameData *in, minrtsFrameData *mask, 
			   float blank)
{
  int reg, i, len;
  float *valIn, *valMask;

  /* sanity checks */
  if (!mask) return;
  if (!in) return;

  /* loop over regions */
  for (reg = 0; reg < in->nregion; reg++)
    {
      len = in->nx[reg] * in->ny[reg]; /* how many elements */
      valIn   = in->data[reg];
      valMask = mask->data[reg];
      for (i = 0; i<len; i++) 
	{
	  if (valMask[i] <= 0.0) valIn[i] = blank;
	}
     } /* end loop over region */
} /* end of minrtsFrameDataBlank */

/**  
  * Checks for saturation, return 0 if OK, >0=>bad
  * Since the A/D units wrap on saturation there is no easy test.
  * Returns number of pixels whose values exceed 32000.0
  * Inputs:
  * \param   in     pointer to minrtsFrameData   input frame
  */
int minrtsFrameDataSaturate (minrtsFrameData *in)
{
  int reg, i, len, bad = 0;
  float *valIn, maxVal;

  /* sanity checks */
  if (!in) return 0;

  /* get Stats if not already there */
  if (in->rms[0]<=0.0) minrtsFrameDataStats(in);

  /* loop over regions */
  for (reg = 0; reg < in->nregion; reg++)
    {
      /* set allowed range of values */
      maxVal = 25000;

      len = in->nx[reg] * in->ny[reg]; /* how many elements */
      valIn = in->data[reg];
      for (i = 0; i<len; i++) 
	{
	  if (valIn[i] > maxVal) bad++;
	}
     } /* end loop over region */
  return bad; /* done */
} /* end of minrtsFrameDataSaturate */
